#!/usr/bin/env python
# Copyright 2024 ARC Centre of Excellence for Climate Extremes
# author: Paola Petrelli <paola.petrelli@utas.edu.au>
# author: Sam Green <sam.green@unsw.edu.au>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This is the ACCESS Model Output Post Processor, derived from the APP4
# originally written for CMIP5 by Peter Uhe and dapted for CMIP6 by Chloe Mackallah
# ( https://doi.org/10.5281/zenodo.7703469 )
#
# last updated 10/10/2024
#
# This file contains a collection of functions to calculate land derived variables
# from ACCESS model output.
# Initial functions' definitions were based on APP4 modified to work with Xarray.
#
# To propose new calculations and/or update to existing ones see documentation:
#
# and open a new issue on github.

from importlib.resources import files as import_files

import numpy as np
import yaml


def extract_tilefrac(tilefrac, tilenum, landfrac=None):
    """Calculates the land fraction of a specific type: crops, grass,
    etc.

    Parameters
    ----------
    ctx : click context
        Includes obj dict with 'cmor' settings, exp attributes
    tilefrac : Xarray DataArray
        variable
    tilenum : Int or [Int]
        the number indicating the tile
    landfrac : Xarray DataArray
        Land fraction variable if None (default) is read from ancil file

    Returns
    -------
    vout : Xarray DataArray
        land fraction of object

    Raises
    ------
    Exception
        tile number must be an integer or list

    """
    pseudo_level = tilefrac.dims[1]
    tilefrac = tilefrac.rename({pseudo_level: "pseudo_level"})
    vout = tilefrac.sel(pseudo_level=tilenum)
    if isinstance(tilenum, int):
        vout = tilefrac.sel(pseudo_level=tilenum)
    elif isinstance(tilenum, list):
        vout = tilefrac.sel(pseudo_level=tilenum).sum(dim="pseudo_level")
    else:
        raise Exception("E: tile number must be an integer or list")
    if landfrac is None:
        # landfrac = get_ancil_var("land_frac", "fld_s03i395")
        raise Exception("E: landfrac not defined")
    vout = vout * landfrac
    return vout.fillna(0)


def calc_topsoil(soilvar):
    """Returns the variable over the first 10cm of soil.

    Parameters
    ----------
    ctx : click context
        Includes obj dict with 'cmor' settings, exp attributes
    soilvar : Xarray DataArray
        Soil moisture over soil levels

    Returns
    -------
    topsoil : Xarray DataArray
        Variable defined on top 10cm of soil

    """
    depth = soilvar.depth
    # find index of bottom depth level including the first 10cm of soil
    maxlev = np.nanargmin(depth.where(depth >= 0.1).values)
    # calculate the fraction of maxlev which falls in first 10cm
    fraction = (0.1 - depth[maxlev - 1]) / (depth[maxlev] - depth[maxlev - 1])
    topsoil = soilvar.isel(depth=slice(0, maxlev)).sum(dim="depth")
    topsoil = topsoil + fraction * soilvar.isel(depth=maxlev)
    return topsoil


def calc_landcover(var, model):
    """Returns land cover fraction variable

    Parameters
    ----------
    ctx : click context obj
        Dictionary including 'cmor' settings and attributes for experiment
    var : list(xarray.DataArray)
        List of input variables to sum
    model: str
        Name of land surface model to retrieve land tiles definitions

    Returns
    -------
    vout : xarray.DataArray
        Land cover faction variable

    """
    fname = import_files("mopdata").joinpath("land_tiles.yaml")
    # data = read_yaml(fname)
    with fname.open(mode="r") as yfile:
        data = yaml.safe_load(yfile)
    vegtype = data[model]
    pseudo_level = var[0].dims[1]
    vout = (var[0] * var[1]).fillna(0)
    vout = vout.rename({pseudo_level: "vegtype"})
    vout["vegtype"] = vegtype
    vout["vegtype"].attrs["units"] = ""
    return vout


def average_tile(var, tilefrac, landfrac=1.0):
    """Returns variable averaged over grid-cell, counting only
    specific tile/s and land fraction when suitable.

    For example: nLitter is nitrogen mass in litter and should be
    calculated only over land fraction and each tile type will have
    different amounts of litter.
    average = sum_over_tiles(N amount on tile * tilefrac) * landfrac

    Parameters
    ----------
    var : Xarray DataArray
        Variable to process defined opver tiles
    tilefrac : Xarray DataArray, optional
        Variable defining tiles' fractions
    landfrac : Xarray DataArray
        Variable defining land fraction (default is 1)

    Returns
    -------
    vout : Xarray DataArray
        averaged input variable

    """
    pseudo_level = var.dims[1]
    vout = var * tilefrac
    vout = vout.sum(dim=pseudo_level)
    vout = vout * landfrac
    return vout



import numpy as np
import xarray as xr


class Supergrid(object):
    # See description in the following document
    # https://gist.github.com/rbeucher/b67c2b461557bc215a70017ea8dd337b

    def __init__(self, supergrid_file):
        self.supergrid_file = supergrid_file
        self.supergrid = xr.open_dataset(supergrid_file)

        # T point locations
        self.xt = self.supergrid["x"][1::2, 1::2]
        self.yt = self.supergrid["y"][1::2, 1::2]
        # Corner point locations
        self.xq = self.supergrid["x"][::2, ::2]
        self.yq = self.supergrid["y"][::2, ::2]
        # U point locations
        self.xu = self.supergrid["x"][1::2, ::2]
        self.yu = self.supergrid["y"][1::2, ::2]
        # V point locations
        self.xv = self.supergrid["x"][::2, 1::2]
        self.yv = self.supergrid["y"][::2, 1::2]

    def h_cells(self):
        self.lat = self.yt.values
        self.lat_bnds = np.zeros((*self.yt.shape, 4))
        self.lat_bnds[..., 0] = self.yq[1:, :-1]  # SW corner
        self.lat_bnds[..., 1] = self.yq[1:, 1:]  # SE corner
        self.lat_bnds[..., 2] = self.yq[:-1, 1:]  # NE corner
        self.lat_bnds[..., 3] = self.yq[:-1, :-1]  # NW corner

        self.lon = (self.xt.values + 360) % 360
        self.xq = (self.xq + 360) % 360
        self.lon_bnds = np.zeros((*self.xt.shape, 4))
        self.lon_bnds[..., 0] = self.xq[1:, :-1]  # SW corner
        self.lon_bnds[..., 1] = self.xq[1:, 1:]  # SE corner
        self.lon_bnds[..., 2] = self.xq[:-1, 1:]  # NE corner
        self.lon_bnds[..., 3] = self.xq[:-1, :-1]  # NW corner

    def q_cells(self):
        self.lat = self.yq.values
        self.lat_bnds = np.zeros((*self.yt.shape, 4))
        self.lat_bnds[..., 0] = self.yt[1:, :-1]  # SW corner
        self.lat_bnds[..., 1] = self.yt[1:, 1:]  # SE corner
        self.lat_bnds[..., 2] = self.yt[:-1, 1:]  # NE corner
        self.lat_bnds[..., 3] = self.yt[:-1, :-1]  # NW corner

        self.lon = (self.xq.values + 360) % 360
        self.xt = (self.xt + 360) % 360
        self.lon_bnds = np.zeros((*self.xq.shape, 4))
        self.lon_bnds[..., 0] = self.xt[1:, :-1]  # SW corner
        self.lon_bnds[..., 1] = self.xt[1:, 1:]  # SE corner
        self.lon_bnds[..., 2] = self.xt[:-1, 1:]  # NE corner
        self.lon_bnds[..., 3] = self.xt[:-1, :-1]  # NW corner



import importlib.resources as resources
from pathlib import Path

import pandas as pd
import pytest
from access_mopper.configurations import ACCESS_ESM16_CMIP6, cmorise

DATA_DIR = Path(__file__).parent / "data"


@pytest.fixture
def model():
    # Create and save the model
    model_instance = ACCESS_ESM16_CMIP6(
        experiment_id="piControl-spinup",
        realization_index="1",
        initialization_index="1",
        physics_index="1",
        forcing_index="1",
        parent_mip_era="no parent",
        parent_activity_id="no parent",
        parent_experiment_id="no parent",
        parent_source_id="no parent",
        parent_variant_label="no parent",
        parent_time_units="no parent",
        branch_method="no parent",
        branch_time_in_parent=0.0,
        branch_time_in_child=0.0,
    )
    model_instance.save_to_file("model.json")
    return model_instance


def test_model_function():
    test_file = DATA_DIR / "esm1-6/atmosphere/aiihca.pa-101909_mon.nc"
    assert test_file.exists(), "Test data file missing!"


def load_filtered_variables(mappings):
    # Load and filter variables from the JSON file
    with resources.files("access_mopper.mappings").joinpath(mappings).open() as f:
        df = pd.read_json(f, orient="index")
    return df.index.tolist()


# @pytest.mark.parametrize("cmor_name", load_filtered_variables("Mappings_CMIP6_Omon.json"))
# def test_cmorise_CMIP6_Omon(model, cmor_name):
#     file_pattern = "/home/romain/PROJECTS/ACCESS-MOPPeR/Test_data//cj877/history/ocn/ocean-2d-sea_level-1-monthly-mean-ym_0326_01.nc"
#     try:
#         mop.cmorise(
#             file_paths=glob.glob(file_pattern),
#             compound_name="Omon."+ cmor_name,
#             reference_time="1850-01-01 00:00:00",
#             cmor_dataset_json="model.json",
#             mip_table="CMIP6_Omon.json"
#         )
#     except Exception as e:
#         pytest.fail(f"Failed processing {cmor_name} with table CMIP6_Omon.json: {e}")


@pytest.mark.parametrize(
    "cmor_name", load_filtered_variables("Mappings_CMIP6_Amon.json")
)
def test_cmorise_CMIP6_Amon(model, cmor_name):
    file_pattern = DATA_DIR / "esm1-6/atmosphere/aiihca.pa-101909_mon.nc"
    try:
        cmorise(
            file_paths=file_pattern,
            compound_name="Amon." + cmor_name,
            cmor_dataset_json="model.json",
            mip_table="CMIP6_Amon.json",
        )
    except Exception as e:
        pytest.fail(f"Failed processing {cmor_name} with table CMIP6_Amon.json: {e}")


@pytest.mark.parametrize(
    "cmor_name", load_filtered_variables("Mappings_CMIP6_Lmon.json")
)
def test_cmorise_CMIP6_Lmon(model, cmor_name):
    file_pattern = DATA_DIR / "esm1-6/atmosphere/aiihca.pa-101909_mon.nc"
    try:
        cmorise(
            file_paths=file_pattern,
            compound_name="Lmon." + cmor_name,
            cmor_dataset_json="model.json",
            mip_table="CMIP6_Lmon.json",
        )
    except Exception as e:
        pytest.fail(f"Failed processing {cmor_name} with table CMIP6_Lmon.json: {e}")


@pytest.mark.parametrize(
    "cmor_name", load_filtered_variables("Mappings_CMIP6_Emon.json")
)
def test_cmorise_CMIP6_Emon(model, cmor_name):
    file_pattern = DATA_DIR / "esm1-6/atmosphere/aiihca.pa-101909_mon.nc"
    try:
        cmorise(
            file_paths=file_pattern,
            compound_name="Emon." + cmor_name,
            cmor_dataset_json="model.json",
            mip_table="CMIP6_Emon.json",
        )
    except Exception as e:
        pytest.fail(f"Failed processing {cmor_name} with table CMIP6_Emon.json: {e}")



#!/usr/bin/env python
# Copyright 2023 ARC Centre of Excellence for Climate Extremes
# author: Paola Petrelli <paola.petrelli@utas.edu.au>
# author: Sam Green <sam.green@unsw.edu.au>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This is the ACCESS Model Output Post Processor, derived from the APP4
# originally written for CMIP5 by Peter Uhe and dapted for CMIP6 by Chloe Mackallah
# ( https://doi.org/10.5281/zenodo.7703469 )
#
# last updated 10/10/2024
#
# This file contains a collection of functions to calculate ocean derived variables
# from ACCESS model output.
# Initial functions' definitions were based on APP4 modified to work with Xarray.
#
# To propose new calculations and/or update to existing ones see documentation:
#
# and open a new issue on github.


import logging
import os

import click
import numpy as np
import xarray as xr
from mopdb.utils import MopException

# from mopper.calc_utils import

# Global Variables
# ----------------------------------------------------------------------

ice_density = 900  # kg/m3
snow_density = 300  # kg/m3

rd = 287.1
cp = 1003.5
p_0 = 100000.0
g_0 = 9.8067  # gravity constant
R_e = 6.378e06
# ----------------------------------------------------------------------


@click.pass_context
def overturn_stream(ctx, varlist, sv=False):
    """Returns ocean overturning mass streamfunction.

    Calculation is:
    sum over the longitudes and cumulative sum over depth for ty_trans var
    then sum these terms to get final values

    Parameters
    ----------
    ctx : click context
        Includes obj dict with 'cmor' settings, exp attributes
    varlist: list( DataArray )
        List of ocean overturning mass streamfunction variables (ty_trans vars)
        From 1-3 if gm and/or submeso are present
    sv: bool
        If True units are sverdrup and they are converted to kg/s
        (default is False)

    Returns
    -------
    stream: DataArray
        The ocean overturning mass streamfunction in kg s-1

    :meta private:
    """
    var_log = logging.getLogger(ctx.obj["var_log"])
    londim = varlist[0].dims[3]
    depdim = varlist[0].dims[1]
    var_log.debug(f"Streamfunct lon, dep dims: {londim}, {depdim}")
    # work out which variables are in list
    var = {"ty": None, "gm": None, "subm": None}
    for v in varlist:
        if "_gm" in v.name:
            var["gm"] = v
        elif "_submeso" in v.name:
            var["subm"] = v
        else:
            var["ty"] = v
    # calculation
    ty_lon = var["ty"].sum(londim)
    stream = ty_lon.cumsum(depdim)
    if var["gm"] is not None:
        stream += var["gm"].sum(londim)
    if var["subm"] is not None:
        stream += var["subm"].sum(londim)
    stream = stream - ty_lon.sum(depdim)
    if sv is True:
        stream = stream * 10**9
    return stream


def ocean_floor(var):
    """Not sure..

    Parameters
    ----------
    var : Xarray dataset
        pot_temp variable

    Returns
    -------
    vout : Xarray dataset
        ocean floor temperature?

        :meta private:
    """
    lv = (~var.isnull()).sum(dim="st_ocean") - 1
    vout = var.take(lv, dim="st_ocean").squeeze()
    return vout


def calc_global_ave_ocean(var, rho_dzt, area_t):
    """Calculate global ocean mass transport.

    Parameters
    ----------
    var : Xarray dataset
        ocean variable
    rho_dzt : Xarray dataset
        density transport
    area_t : Xarray dataset
        area transport

    Returns
    -------
    vnew : Xarray dataset
        global ocean mass transport

    :meta private:
    """
    mass = rho_dzt * area_t
    # PP would be better to get the correct dimension from variable and use them
    # rather than try and except

    try:
        vnew = var.weighted(mass).mean(
            dim=("st_ocean", "yt_ocean", "xt_ocean"), skipna=True
        )
    except Exception as e:
        vnew = var.weighted(mass[:, 0, :, :]).mean(dim=("x", "y"), skipna=True)

    return vnew


@click.pass_context
def calc_global_ave_ocean(ctx, var, rho_dzt):
    """Returns global average ocean temperature

    NB needs checking

    Parameters
    ----------
    ctx : click context
        Includes obj dict with 'cmor' settings, exp attributes
    var : xarray.DataArray
        Input variable
    rho_dzt: Xarray DataArray
        sea_water_mass_per_unit_area dimensions: (time, depth, lat, lon)

    Returns
    -------
    vnew : xarray.DataArray
        output

    :meta private:
    """
    fname = f"{ctx.obj['ancils_path']}/{ctx.obj['grid_ocean']}"
    ds = xr.open_dataset(fname)
    area_t = ds["area_t"].reindex_like(rho_dzt, method="nearest")
    mass = rho_dzt * area_t
    try:
        vnew = np.average(var, axis=(1, 2, 3), weights=mass)
    except Exception as e:
        vnew = np.average(var, axis=(1, 2), weights=mass[:, 0, :, :])

    return vnew


@click.pass_context
def calc_overt(ctx, varlist, sv=False):
    """Returns overturning mass streamfunction variable

    Parameters
    ----------
    ctx : click context
        Includes obj dict with 'cmor' settings, exp attributes
    varlist: list( DataArray )
        List of ocean transport variables (ty_trans vars)
        From 1-3 if gm and/or submeso are present
    sv: bool
        If True units are sverdrup and they are converted to kg/s
        (default is False)
    varlist: list( DataArray )
        transport components to use to calculate streamfunction

    Returns
    -------
    overt: DataArray
        overturning mass streamfunction (time, basin, depth, gridlat) variable

    """
    var_log = logging.getLogger(ctx.obj["var_log"])
    var1 = varlist[0]
    vlat, vlon = var1.dims[2:]
    mask = get_basin_mask(vlat, vlon)
    mlat = mask.dims[0]
    mlon = mask.dims[1]
    if [mlat, mlon] != [vlat, vlon]:
        # if mask uses different lat/lon interp mask to var dimesnions
        # mask = mask.sel(mlat=vlat, mlon=vlon, method="nearest")
        mask = mask.sel(**{mlat: var1[vlat], mlon: var1[vlon]}, method="nearest")
    var_log.debug(f"Basin mask: {mask}")
    # first calculate for global ocean
    glb = overturn_stream(varlist)
    # atlantic and arctic basin have mask values 2 and 4 #TODO double check this
    var_masked = [v.where(mask.isin([2, 4]), 0) for v in varlist]
    atl = overturn_stream(var_masked)
    # Indian and Pacific basin are given by mask values 3 and 5 #TODO double check this
    var_masked = [v.where(mask.isin([3, 5]), 0) for v in varlist]
    ind = overturn_stream(var_masked)
    # now add basin dimension to resulting array
    glb = glb.expand_dims(dim={"basin": ["global_ocean"]}, axis=1)
    atl = atl.expand_dims(dim={"basin": ["atlantic_arctic_ocean"]}, axis=1)
    ind = ind.expand_dims(dim={"basin": ["indian_pacific_ocean"]}, axis=1)
    overt = xr.concat([atl, ind, glb], dim="basin", coords="minimal")
    if ctx.obj["variable_id"][:5] == "msfty":
        overt = overt.rename({vlat: "gridlat"})
    overt["basin"].attrs["units"] = ""
    return overt


@click.pass_context
def get_areacello(ctx, area_t=None):
    """Returns areacello

    Parameters
    ----------
    ctx : click context
        Includes obj dict with 'cmor' settings, exp attributes
    area_t: DataArray
        area of t-cells (default None then is read from ancil file)

    Returns
    -------
    areacello: DataArray
        areacello variable

    """
    fname = f"{ctx.obj['ancils_path']}/{ctx.obj['grid_ocean']}"
    ds = xr.open_dataset(fname)
    if area_t is None:
        area_t = ds.area_t
    areacello = xr.where(ds.ht.isnull(), 0, ds.area_t)
    return areacello


@click.pass_context
def get_basin_mask(ctx, lat, lon):
    """Returns first level of basin mask from lsmask ancil file.

    Lat, lon are used to work out which mask to use tt, uu, ut, tu
        where t/u refer to t/c cell, for x/y axis
    For example ut stands for c-cell lon and t-cell lat

    Parameters
    ----------
    ctx : click context
        Includes obj dict with 'cmor' settings, exp attributes
    lat: str
        latitude coordinate name
    lon: str
        longitude coordinate name

    Returns
    -------
    basin_mask: DataArray
        basin_mask(lat,lon)

    :meta private:
    """
    var_log = logging.getLogger(ctx.obj["var_log"])
    coords = ["t", "t"]
    if "xu" in lon:
        coords[0] = "u"
    elif "yu" in lat:
        coords[1] = "u"
    fname = f"{ctx.obj['ancils_path']}/{ctx.obj['mask_ocean']}"
    if os.path.isfile(fname):
        ds = xr.open_dataset(fname)
    else:
        var_log.error(f"Ocean mask file {fname} doesn't exists")
        raise MopException(f"Ocean mask file {fname} doesn't exists")
    # based on coords select mask
    mask = f"mask_{''.join(coords)}cell"
    basin_mask = ds[mask].isel(st_ocean=0).fillna(0)
    return basin_mask



#!/usr/bin/env python
# Copyright 2024 ARC Centre of Excellence for Climate Extremes
# author: Paola Petrelli <paola.petrelli@utas.edu.au>
# author: Sam Green <sam.green@unsw.edu.au>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This is the ACCESS Model Output Post Processor, derived from the APP4
# originally written for CMIP5 by Peter Uhe and dapted for CMIP6 by Chloe Mackallah
# ( https://doi.org/10.5281/zenodo.7703469 )
#
# last updated 10/10/2024
#
# This file contains a collection of functions to calculate seaice derived variables
# from ACCESS model output.
# Initial functions' definitions were based on APP4 modified to work with Xarray.
#
# To propose new calculations and/or update to existing ones see documentation:
#
# and open a new issue on github.


from importlib.resources import files as import_files

import click
import numpy as np
import xarray as xr
from mopdb.utils import MopException, read_yaml

# Global Variables
# ----------------------------------------------------------------------

ice_density = 900  # kg/m3
snow_density = 300  # kg/m3

rd = 287.1
cp = 1003.5
p_0 = 100000.0
g_0 = 9.8067  # gravity constant
R_e = 6.378e06
# ----------------------------------------------------------------------


class IceTransportCalculations:
    """
    Functions to calculate mass transports.

    Parameters
    ----------

    Returns
    -------
    transports : Xarray DataArray
        mass transport array

    :meta private:
    """

    @click.pass_context
    def __init__(self, ctx):
        fname = import_files("mopdata").joinpath("transport_lines.yaml")
        self.yaml_data = read_yaml(fname)["lines"]

        self.gridfile = xr.open_dataset(
            f"{ctx.obj['ancils_path']}/" + f"{ctx.obj['grid_ice']}"
        )
        self.lines = self.yaml_data["sea_lines"]
        self.ice_lines = self.yaml_data["ice_lines"]

    def __del__(self):
        self.gridfile.close()

    def get_grid_cell_length(self, xy):
        """
         Select the hun or hue variable from the opened gridfile depending on whether
         x or y are passed in.


         Parameters
         ----------
         xy : string
             axis name

         Returns
         -------
         L : Xarray dataset
             hun or hue variable

        :meta private:
        """
        if xy == "y":
            L = self.gridfile.hun / 100  # grid cell length in m (from cm)
        elif xy == "x":
            L = self.gridfile.hue / 100  # grid cell length in m (from cm)
        else:
            raise Exception("""Need to supply value either 'x' or 'y'
                            for ice Transports""")

        return L

    def transAcrossLine(self, var, i_start, i_end, j_start, j_end):
        """Calculates the mass trasport across a line either
         i_start=i_end and the line goes from j_start to j_end or
         j_start=j_end and the line goes from i_start to i_end.
         var is either the x or y mass transport depending on the line.


         Parameters
         ----------
         var : DataArray
             variable extracted from Xarray dataset
         i_start: int
             xt_ocean axis position
         i_end: int
             xt_ocean axis position
         j_start: int
             yu_ocean axis position
         j_end: int
             yu_ocean axis position


         Returns
         -------
         transports : DataArray

        :meta private:
        """
        # PP is it possible to generalise this? as I'm sure I have to do the same in main
        # code to work out correct coordinates
        if "yt_ocean" in var:
            y_ocean = "yt_ocean"
            x_ocean = "xu_ocean"
        else:
            y_ocean = "yu_ocean"
            x_ocean = "xt_ocean"

        # could we try to make this a sel lat lon values?
        if i_start == i_end or j_start == j_end:
            try:
                # sum each axis apart from time (3d)
                # trans = var.isel(yu_ocean=slice(271, 271+1), xt_ocean=slice(292, 300+1))
                trans = var[..., j_start : j_end + 1, i_start : i_end + 1].sum(
                    dim=["st_ocean", f"{y_ocean}", f"{x_ocean}"]
                )  # 4D
            except Exception as e:
                trans = var[..., j_start : j_end + 1, i_start : i_end + 1].sum(
                    dim=[f"{y_ocean}", f"{x_ocean}"]
                )  # 3D

            return trans
        else:
            raise Exception("""ERROR: Transport across a line needs to
                be calculated for a single value of i or j""")

    def lineTransports(self, tx_trans, ty_trans):
        """
         Calculates the mass transports across the ocn straits.


         Parameters
         ----------
         tx_trans : DataArray
             variable extracted from Xarray dataset
         ty_trans: DataArray
             variable extracted from Xarray dataset


         Returns
         -------
         trans : Datarray

        :meta private:
        """
        # PP these are all hardcoded need to change this to be dependent on grid!!!
        # initialise array
        transports = np.zeros([len(tx_trans.time), len(self.lines)])

        # 0 barents opening
        transports[:, 0] = self.transAcrossLine(ty_trans, 292, 300, 271, 271)
        transports[:, 0] += self.transAcrossLine(tx_trans, 300, 300, 260, 271)

        # 1 bering strait
        transports[:, 1] = self.transAcrossLine(ty_trans, 110, 111, 246, 246)

        # 2 canadian archipelago
        transports[:, 2] = self.transAcrossLine(ty_trans, 206, 212, 285, 285)
        transports[:, 2] += self.transAcrossLine(tx_trans, 235, 235, 287, 288)

        # 3 denmark strait
        transports[:, 3] = self.transAcrossLine(tx_trans, 249, 249, 248, 251)
        transports[:, 3] += self.transAcrossLine(ty_trans, 250, 255, 247, 247)

        # 4 drake passage
        transports[:, 4] = self.transAcrossLine(tx_trans, 212, 212, 32, 49)

        # 5 english channel
        # Is unresolved by the access model

        # 6 pacific equatorial undercurrent
        # specified down to 350m not the whole depth
        tx_trans_ma = tx_trans.where(tx_trans[:, 0:25, :] >= 0)
        transports[:, 6] = self.transAcrossLine(tx_trans_ma, 124, 124, 128, 145)

        # 7 faroe scotland channel
        transports[:, 7] = self.transAcrossLine(ty_trans, 273, 274, 238, 238)
        transports[:, 7] += self.transAcrossLine(tx_trans, 274, 274, 232, 238)

        # 8 florida bahamas strait
        transports[:, 8] = self.transAcrossLine(ty_trans, 200, 205, 192, 192)

        # 9 fram strait
        transports[:, 9] = self.transAcrossLine(tx_trans, 267, 267, 279, 279)
        transports[:, 9] += self.transAcrossLine(ty_trans, 268, 284, 278, 278)

        # 10 iceland faroe channel
        transports[:, 10] = self.transAcrossLine(ty_trans, 266, 268, 243, 243)
        transports[:, 10] += self.transAcrossLine(tx_trans, 268, 268, 240, 243)
        transports[:, 10] += self.transAcrossLine(ty_trans, 269, 272, 239, 239)
        transports[:, 10] += self.transAcrossLine(tx_trans, 272, 272, 239, 239)

        # 11 indonesian throughflow
        transports[:, 11] = self.transAcrossLine(tx_trans, 31, 31, 117, 127)
        transports[:, 11] += self.transAcrossLine(ty_trans, 35, 36, 110, 110)
        transports[:, 11] += self.transAcrossLine(ty_trans, 43, 44, 110, 110)
        transports[:, 11] += self.transAcrossLine(tx_trans, 46, 46, 111, 112)
        transports[:, 11] += self.transAcrossLine(ty_trans, 47, 57, 113, 113)

        # 12 mozambique channel
        transports[:, 12] = self.transAcrossLine(ty_trans, 320, 323, 91, 91)

        # 13 taiwan luzon straits
        transports[:, 13] = self.transAcrossLine(ty_trans, 38, 39, 190, 190)
        transports[:, 13] += self.transAcrossLine(tx_trans, 40, 40, 184, 188)

        # 14 windward passage
        transports[:, 14] = self.transAcrossLine(ty_trans, 205, 206, 185, 185)

        return transports

    def iceTransport(self, ice_thickness, vel, xy):
        """
        Calculate ice mass transport.


        Parameters
        ----------
        ice_thickness : DataArray
            variable extracted from Xarray dataset
        vel: DataArray
            variable extracted from Xarray dataset
        xy: str
            variable extracted from Xarray dataset


        Returns
        -------
        ice_mass : DataArray


        :meta private:
        """
        L = self.gridfile(xy)
        ice_mass = ice_density * ice_thickness * vel * L

        return ice_mass

    def snowTransport(self, snow_thickness, vel, xy):
        """
        Calculate snow mass transport.


        Parameters
        ----------
        snow_thickness : DataArray
            variable extracted from Xarray dataset
        vel: DataArray
            variable extracted from Xarray dataset
        xy: str
            variable extracted from Xarray dataset


        Returns
        -------
        snow_mass : DataArray

        :meta private:
        """
        L = self.gridfile(xy)
        snow_mass = snow_density * snow_thickness * vel * L

        return snow_mass

    def iceareaTransport(self, ice_fraction, vel, xy):
        """
        Calculate ice area transport.


        Parameters
        ----------
        ice_fraction : DataArray
            variable extracted from Xarray dataset
        vel: DataArray
            variable extracted from Xarray dataset
        xy: str
            variable extracted from Xarray dataset


        Returns
        -------
        ice_area : DataArray

        :meta private:
        """
        L = self.gridfile(xy)
        ice_area = ice_fraction * vel * L

        return ice_area

    def fill_transports(self, tx_trans, ty_trans):
        """
        Calculates the mass transports across the ice straits.


        Parameters
        ----------
        tx_trans : DataArray
            variable extracted from Xarray dataset
        ty_trans: DataArray
            variable extracted from Xarray dataset


        Returns
        -------
        transports : DataArray

        :meta private:
        """
        transports = np.zeros([len(tx_trans.time), len(self.lines)])

        # PP these are all hardcoded need to change this to be dependent on grid!!!
        # 0 fram strait
        transports[:, 0] = self.transAcrossLine(tx_trans, 267, 267, 279, 279)
        transports[:, 0] += self.transAcrossLine(ty_trans, 268, 284, 278, 278)

        # 1 canadian archipelago
        transports[:, 1] = self.transAcrossLine(ty_trans, 206, 212, 285, 285)
        transports[:, 1] += self.transAcrossLine(tx_trans, 235, 235, 287, 288)

        # 2 barents opening
        transports[:, 2] = self.transAcrossLine(ty_trans, 292, 300, 271, 271)
        transports[:, 2] += self.transAcrossLine(tx_trans, 300, 300, 260, 271)

        # 3 bering strait
        transports[:, 3] = self.transAcrossLine(ty_trans, 110, 111, 246, 246)

        return transports

    def icelineTransports(self, ice_thickness, velx, vely):
        """
        Calculates the ice mass transport across the straits


        Parameters
        ----------
        ice_thickness : DataArray
            variable extracted from Xarray dataset
        velx : DataArray
            variable extracted from Xarray dataset
        vely: DataArray
            variable extracted from Xarray dataset


        Returns
        -------
        transports : DataArray

        :meta private:
        """

        tx_trans = self.iceTransport(ice_thickness, velx, "x").fillna(0)
        ty_trans = self.iceTransport(ice_thickness, vely, "y").fillna(0)
        transports = self.fill_transports(tx_trans, ty_trans)

        return transports

    def snowlineTransports(self, snow_thickness, velx, vely):
        """
        Calculates the Snow mass transport across the straits


        Parameters
        ----------
        snow_thickness : DataArray
            variable extracted from Xarray dataset
        velx : DataArray
            variable extracted from Xarray dataset
        vely: DataArray
            variable extracted from Xarray dataset


        Returns
        -------
        transports : DataArray

        :meta private:
        """
        tx_trans = self.snowTransport(snow_thickness, velx, "x").fillna(0)
        ty_trans = self.snowTransport(snow_thickness, vely, "y").fillna(0)
        transports = self.fill_transports(tx_trans, ty_trans)

        return transports

    def icearealineTransports(self, ice_fraction, velx, vely):
        """
        Calculates the ice are transport across the straits


        Parameters
        ----------
        ice_fraction : DataArray
            variable extracted from Xarray dataset
        velx : DataArray
            variable extracted from Xarray dataset
        vely: DataArray
            variable extracted from Xarray dataset


        Returns
        -------
        transports : DataArray

        :meta private:
        """
        tx_trans = self.iceareaTransport(ice_fraction, velx, "x").fillna(0)
        ty_trans = self.iceareaTransport(ice_fraction, vely, "y").fillna(0)
        transports = self.fill_transports(tx_trans, ty_trans)

        return transports

    def msftbarot(self, psiu, tx_trans):
        """
        Calculates the drake trans


        Parameters
        ----------
        psiu : DataArray
            variable extracted from Xarray dataset
        tx_trans : DataArray
            variable extracted from Xarray dataset


        Returns
        -------
        psiu : DataArray

        :meta private:
        """
        # PP these are all hardcoded need to change this to be dependent on grid!!!
        drake_trans = self.transAcrossLine(tx_trans, 212, 212, 32, 49)
        # loop over times
        for i, trans in enumerate(drake_trans):
            # offset psiu by the drake passage transport at that time
            psiu[i, :] = psiu[i, :] + trans
        return psiu


class SeaIceCalculations:
    """
    Functions to calculate mass transports.

    Parameters
    ----------

    Returns
    -------
    transports : Xarray DataArray
        mass transport array

    :meta private:
    """

    @click.pass_context
    def __init__(self, ctx):
        fname = import_files("mopdata").joinpath("transport_lines.yaml")
        self.yaml_data = read_yaml(fname)["lines"]

        self.gridfile = xr.open_dataset(
            f"{ctx.obj['ancil_path']}/" + f"{ctx.obj['grid_ice']}"
        )
        self.lines = self.yaml_data["sea_lines"]
        self.ice_lines = self.yaml_data["ice_lines"]

    def __del__(self):
        self.gridfile.close()


def calc_hemi_seaice(invar, carea, hemi, extent=False):
    """Calculate seaice properties (volume, area and extent) over
    hemisphere.

    Parameters
    ----------
    invar : Xarray DataArray
        Variable to process, either fraction (aice) or volume (hi)
    carea : Xarray DataArray
        Grid cell area
    hemi : str
            Assigning the hemisphere to calculate, either 'north' or'south'.
    extent : bool
        True if calculaitng extent: tarea as extent and var (aice) as filter.

    Returns
    -------
    vout : xarray DataArray
        Sum of property over selected hemisphere

    """
    vlat = invar.dims[1]
    # if calculating extent sum carea and aice is used as filter
    # with volume and area invar is multiplied by carea first
    if extent:
        var = tarea.where(invar <= 1.0 and invar >= 0.15, drop=True)
    else:
        var = invar * tarea
    if hemi == "north":
        var = var.sel(vlat >= 0.0)
    elif hemi == "south":
        var = var.sel(vlat < 0.0)
    else:
        mop_log.error(f"invalid hemisphere: {hemi}")
        raise MopException(f"invalid hemisphere: {hemi}")
    vout = var.sum()
    return vout


def maskSeaIce(var, aice):
    """Mask seaice.

    Parameters
    ----------
    var : Xarray dataset
        seaice variable
    aice : Xarray dataset
        seaice fraction

    Returns
    -------
    vout : Xarray dataset
        masked seaice variable

    """
    vout = var.where(aice != 0)
    return vout


def sithick(hi, aice):
    """Calculate seaice thickness.

    Parameters
    ----------
    hi : Xarray dataset
        seaice thickness
    aice : Xarray dataset
        seaice fraction

    Returns
    -------
    vout : Xarray dataset
        seaice thickness

    :meta private:
    """
    aice = aice.where(aice > 1e-3, drop=True)
    vout = hi / aice
    return vout


def sisnconc(sisnthick):
    """Calculate seas ice?

    Parameters
    ----------
    sisnthick : Xarray dataset

    Returns
    -------
    vout : Xarray dataset

    :meta private:
    """
    vout = 1 - np.exp(-0.2 * 330 * sisnthick)
    return vout



import importlib.resources as resources
import json
from dataclasses import dataclass, field

import yaml

from access_mopper import _creator


@dataclass
class ACCESS_Dataset:
    # General attributes
    Conventions: str = "CF-1.7, ACDD-1.3"
    comment: str = (
        "post-processed using ACCESS-MOPPeR v2, please contact ACCESS-NRI for questions"
    )
    license: str = "https://creativecommons.org/licenses/by/4.0/"

    # General information for ACCESS models
    source_id: str = ""
    source: str = ""
    keywords: str = ""
    references: str = ""
    forcing: str = ""
    calendar: str = ""
    grid: str = ""
    grid_label: str = ""
    nominal_resolution: str = ""
    parent: bool = None
    tracking_id_prefix: str = ""


@dataclass
class ACCESS_Experiment(ACCESS_Dataset):
    title: str = ""
    exp_description: str = ""
    product_version: str = ""
    date_created: str = ""
    time_coverage_start: str = ""
    time_coverage_end: str = ""
    outpath: str = "MOPPeR_outputs"

    creator_name: str = field(default_factory=lambda: _creator.creator_name)
    creator_email: str = field(default_factory=lambda: _creator.creator_email)
    creator_url: str = field(default_factory=lambda: _creator.creator_url)
    organisation: str = field(default_factory=lambda: _creator.organisation)

    mapping_file: str = ""

    def initialise(self, access_configuration):
        with (
            resources.files("access_mopper")
            .joinpath("ACCESS_configurations.yml")
            .open() as f
        ):
            yaml_data = yaml.safe_load(f)
        attributes = yaml_data[access_configuration]

        for key, value in attributes.items():
            if hasattr(self, key):
                setattr(self, key, value)

    # Method to save the instance data to a file
    def save_to_file(self, file_path: str):
        # Convert the dataclass to a dictionary and then to a JSON string
        json_data = json.dumps(self.__dict__, indent=4)

        # Write the JSON string to a file
        with open(file_path, "w") as f:
            f.write(json_data)
        print(f"Data saved to {file_path}")

    @classmethod
    def get_mapping(cls, compound_name):
        mip_name, cmor_name = compound_name.split(".")
        filename = f"{cls.mapping_file_prefix}{mip_name}.json"
        # Use importlib.resources to access the file
        with (
            resources.files("access_mopper.mappings")
            .joinpath(filename)
            .open("r") as file
        ):
            data = json.load(file)
        return data[cmor_name]

    @classmethod
    def mapping_info(cls, compound_name):
        """
        Prints the mapping information for a given compound name in a notebook-friendly format.

        Args:
            compound_name (str): The compound name in the format "MIP_table.CMOR_variable".
        """
        from IPython.display import Markdown, display

        # Get the mapping data
        mapping = cls.get_mapping(compound_name)

        # Extract relevant information
        mip_table, cmor_name = compound_name.split(".")
        cf_name = mapping.get("CF standard name", "N/A")
        model_variables = mapping.get("model_variables", [])
        formula = mapping.get("calculation", {}).get("formula", "N/A")

        # Format the output as Markdown
        output = f"""
    ### Mapping Information for `{compound_name}`
    - **Compound Name**: `{compound_name}`
    - **CF Standard Name**: `{cf_name}`
    - **Required Variables**: `{", ".join(model_variables)}`
    - **Formula**: `{formula}`
        """

        # Display the Markdown content
        display(Markdown(output.strip()))


@dataclass
class CMIP6_Experiment(ACCESS_Experiment):
    Conventions: str = ""
    institution_id: str = ""
    source_id: str = ""
    source_type: str = "AOGCM"
    experiment_id: str = ""
    activity_id: str = "CMIP"
    realization_index: str = ""
    initialization_index: str = ""
    physics_index: str = ""
    forcing_index: str = ""
    tracking_prefix: str = "hdl:21.14100"
    parent_experiment_id: str = "none"
    parent_activity_id: str = "none"
    parent_source_id: str = "none"
    parent_variant_label: str = "none"
    sub_experiment: str = "none"
    sub_experiment_id: str = "none"
    branch_method: str = "none"
    branch_time_in_child: str = ""
    branch_time_in_parent: str = ""
    _controlled_vocabulary_file: str = "CMIP6_CV.json"
    _AXIS_ENTRY_FILE: str = "CMIP6_coordinate.json"
    _FORMULA_VAR_FILE: str = "CMIP6_formula_terms.json"
    _cmip6_option: str = "CMIP6"
    mip_era: str = "CMIP6"
    parent_mip_era: str = "CMIP6"
    parent_time_units: str = ""
    _history_template: str = "%s ;rewrote data to be consistent with <activity_id> for variable <variable_id> found in table <table_id>."
    output_path_template: str = "<mip_era><activity_id><institution_id><source_id><experiment_id><_member_id><table><variable_id><grid_label><version>"
    output_file_template: str = (
        "<variable_id><table><source_id><experiment_id><_member_id><grid_label>"
    )
    license: str = "CMIP6 model data produced by CSIRO is licensed under a Creative Commons Attribution-ShareAlike 4.0 International License (https://creativecommons.org/licenses/). Consult https://pcmdi.llnl.gov/CMIP6/TermsOfUse for terms of use governing CMIP6 output, including citation requirements and proper acknowledgment.  Further information about this data, including some limitations, can be found via the further_info_url (recorded as a global attribute in this file). The data producers and data providers make no warranty, either express or implied, including, but not limited to, warranties of merchantability and fitness for a particular purpose. All liabilities arising from the supply of the information (including any liability arising in negligence) are excluded to the fullest extent permitted by law."



import operator
import os
from dataclasses import dataclass

import cmor
import numpy as np
import xarray as xr

from .calc_atmos import level_to_height
from .calc_land import average_tile, calc_landcover, calc_topsoil, extract_tilefrac
from .dataclasses import CMIP6_Experiment
from .ocean_supergrid import Supergrid

# Supported operators
OPERATORS = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "/": operator.truediv,
    "**": operator.pow,
}


@dataclass
class ACCESS_ESM16_CMIP6(CMIP6_Experiment):
    mapping_file_prefix: str = "Mappings_CMIP6_"
    supergrid_file = ""
    _supergrid = None

    def __post_init__(self):
        self.initialise("ACCESS-ESM1-5")

    @property
    def supergrid(self):
        return self._supergrid

    @supergrid.setter
    def supergrid(self, path):
        self._supergrid = Supergrid(path)
        self.supergrid_file = path

    def cmorise(self, file_paths, compound_name, cmor_dataset_json, mip_table):
        cmor_name = compound_name.split(".")[1]

        # Open the matching files with xarray
        ds = xr.open_mfdataset(file_paths, combine="by_coords", decode_times=False)

        # Extract required variables and coordinates
        mapping = self.get_mapping(compound_name=compound_name)

        if mapping["calculation"]["type"] == "direct":
            access_var = mapping["calculation"]["formula"]
            variable_units = mapping["units"]
            positive = mapping["positive"]
            var = ds[access_var]
        else:
            access_vars = {var: ds[var] for var in mapping["model_variables"]}
            formula = mapping["calculation"]["formula"]
            variable_units = mapping["units"]
            positive = mapping["positive"]
            custom_functions = {
                "level_to_height": level_to_height,
                "extract_tilefrac": extract_tilefrac,
                "calc_landcover": calc_landcover,
                "calc_topsoil": calc_topsoil,
                "average_tile": average_tile,
            }
            try:
                context = {**access_vars, **OPERATORS, **custom_functions}
                var = eval(formula, {"__builtins__": None}, context)
            except Exception as e:
                raise ValueError(f"Error evaluating formula '{formula}': {e}")

        dim_mapping = mapping["dimensions"]
        axes = {dim_mapping.get(axis, axis): axis for axis in var.dims}

        data = var.values
        lat_axis = axes.pop("latitude")
        lat = ds[lat_axis].values
        lat_bnds = ds[ds[lat_axis].attrs["bounds"]].values
        lon_axis = axes.pop("longitude")
        lon = ds[lon_axis].values
        lon_bnds = ds[ds[lon_axis].attrs["bounds"]].values

        # Convert time to numeric values
        time_axis = axes.pop("time")
        time_numeric = ds[time_axis].values
        time_units = ds[time_axis].attrs["units"]
        time_bnds = ds[ds[time_axis].attrs["bounds"]].values
        # TODO: Check that the calendar is the same than the one defined in the model.json
        # Convert if not.
        # calendar = ds[time_axis].attrs["calendar"]

        # CMOR setup
        ipth = "Test"
        cmor.setup(
            inpath=ipth,
            set_verbosity=cmor.CMOR_NORMAL,
            netcdf_file_action=cmor.CMOR_REPLACE,
        )

        cmor.dataset_json(cmor_dataset_json)
        current_dir = os.path.dirname(os.path.abspath(__file__))
        mip_table = os.path.join(current_dir, "cmor_tables", mip_table)
        cmor.load_table(mip_table)

        cmor_axes = []
        # Define CMOR axes
        cmorLat = cmor.axis(
            "latitude", coord_vals=lat, cell_bounds=lat_bnds, units="degrees_north"
        )
        cmor_axes.append(cmorLat)
        cmorLon = cmor.axis(
            "longitude", coord_vals=lon, cell_bounds=lon_bnds, units="degrees_east"
        )
        cmor_axes.append(cmorLon)
        cmorTime = cmor.axis(
            "time", coord_vals=time_numeric, cell_bounds=time_bnds, units=time_units
        )
        cmor_axes.append(cmorTime)

        if axes:
            for axis, dim in axes.items():
                coord_vals = var[dim].values
                try:
                    cell_bounds = var[var[dim].attrs["bounds"]].values
                except KeyError:
                    cell_bounds = None
                axis_units = var[dim].attrs["units"]
                cmor_axis = cmor.axis(
                    axis,
                    coord_vals=coord_vals,
                    cell_bounds=cell_bounds,
                    units=axis_units,
                )
                cmor_axes.append(cmor_axis)

        # Define CMOR variable
        cmorVar = cmor.variable(cmor_name, variable_units, cmor_axes, positive=positive)

        # Write data to CMOR
        cmor.write(cmorVar, data, ntimes_passed=len(time_numeric))

        # Finalize and save the file
        filename = cmor.close(cmorVar, file_name=True)
        print("Stored in:", filename)

        cmor.close()

    def cmorise_ocean(self, file_paths, compound_name, cmor_dataset_json, mip_table):
        _, cmor_name = compound_name.split(".")

        # Open the matching files with xarray
        ds = xr.open_mfdataset(file_paths, combine="by_coords", decode_times=False)

        # Extract required variables and coordinates
        mapping = self.get_mapping(compound_name=compound_name)

        if mapping["calculation"]["type"] == "direct":
            access_var = mapping["calculation"]["formula"]
            variable_units = mapping["units"]
            positive = mapping["positive"]
            var = ds[access_var]
        else:
            raise ValueError("Not implemented")

        dim_mapping = mapping["dimensions"]
        _ = (
            self.supergrid.h_cells()
            if "xt_ocean" in var.dims
            else self.supergrid.q_cells()
        )
        axes = {dim_mapping.get(axis, axis): axis for axis in var.dims}

        i_axis = axes.pop("longitude")
        i_axis = ds[i_axis].values
        j_axis = axes.pop("latitude")
        j_axis = ds[j_axis].values
        x = np.arange(i_axis.size, dtype="float")
        x_bnds = np.array([[x_ - 0.5, x_ + 0.5] for x_ in x])
        y = np.arange(j_axis.size, dtype="float")
        y_bnds = np.array([[y_ - 0.5, y_ + 0.5] for y_ in y])

        data = var.values
        lat = self.supergrid.lat
        lat_bnds = self.supergrid.lat_bnds

        lon = self.supergrid.lon
        lon_bnds = self.supergrid.lon_bnds

        # Convert time to numeric values
        time_axis = axes.pop("time")
        time_numeric = ds[time_axis].values
        time_units = ds[time_axis].attrs["units"]
        time_bnds = ds[ds[time_axis].attrs["bounds"]].values
        # TODO: Check that the calendar is the same than the one defined in the model.json
        # Convert if not.
        # calendar = ds[time_axis].attrs["calendar"]

        # CMOR setup
        ipth = "Test"
        cmor.setup(
            inpath=ipth,
            set_verbosity=cmor.CMOR_NORMAL,
            netcdf_file_action=cmor.CMOR_REPLACE,
        )

        cmor.dataset_json(cmor_dataset_json)
        current_dir = os.path.dirname(os.path.abspath(__file__))
        grid_table_id = cmor.load_table(
            os.path.join(current_dir, "cmor_tables", "CMIP6_grids.json")
        )
        cmor.set_table(grid_table_id)

        cmor_axes = []
        # Define CMOR axes
        yaxis_id = cmor.axis(
            table_entry="j_index", units="1", coord_vals=y, cell_bounds=y_bnds
        )
        xaxis_id = cmor.axis(
            table_entry="i_index", units="1", coord_vals=x, cell_bounds=x_bnds
        )

        grid_id = cmor.grid(
            axis_ids=np.array([yaxis_id, xaxis_id]),
            latitude=lat,
            longitude=lon,
            latitude_vertices=lat_bnds,
            longitude_vertices=lon_bnds,
        )
        cmor_axes.append(grid_id)

        # Now, load the Omon table to set up the time axis and variable
        current_dir = os.path.dirname(os.path.abspath(__file__))
        mip_table = os.path.join(current_dir, "cmor_tables", mip_table)
        omon_table_id = cmor.load_table(mip_table)
        cmor.set_table(omon_table_id)

        cmorTime = cmor.axis(
            "time", coord_vals=time_numeric, cell_bounds=time_bnds, units=time_units
        )
        cmor_axes.append(cmorTime)

        if axes:
            for axis, dim in axes.items():
                coord_vals = var[dim].values
                try:
                    cell_bounds = var[var[dim].attrs["bounds"]].values
                except KeyError:
                    cell_bounds = None
                axis_units = var[dim].attrs["units"]
                cmor_axis = cmor.axis(
                    axis,
                    coord_vals=coord_vals,
                    cell_bounds=cell_bounds,
                    units=axis_units,
                )
                cmor_axes.append(cmor_axis)

        # Define CMOR variable
        cmorVar = cmor.variable(cmor_name, variable_units, cmor_axes, positive=positive)

        # Write data to CMOR
        data = np.moveaxis(data, 0, -1)
        cmor.write(cmorVar, data, ntimes_passed=len(time_numeric))

        # Finalize and save the file
        filename = cmor.close(cmorVar, file_name=True)
        print("Stored in:", filename)

        cmor.close()


@dataclass
class ACCESS_OM3_CMIP6(CMIP6_Experiment):
    mapping_file_prefix: str = "Mappings_OM3_"
    supergrid_file = ""
    _supergrid = None

    def __post_init__(self):
        self.initialise("ACCESS-OM3")

    @property
    def supergrid(self):
        return self._supergrid

    @supergrid.setter
    def supergrid(self, path):
        self._supergrid = Supergrid(path)
        self.supergrid_file = path

    def cmorise(self, file_paths, compound_name, cmor_dataset_json, mip_table):
        _, cmor_name = compound_name.split(".")

        # Open the matching files with xarray
        ds = xr.open_mfdataset(file_paths, combine="by_coords", decode_times=False)

        # Extract required variables and coordinates
        mapping = self.get_mapping(compound_name=compound_name)

        if mapping["calculation"]["type"] == "direct":
            access_var = mapping["calculation"]["formula"]
            variable_units = mapping["units"]
            positive = mapping["positive"]
            var = ds[access_var]
        else:
            raise ValueError("Not implemented")

        dim_mapping = mapping["dimensions"]
        _ = self.supergrid.h_cells() if "xh" in var.dims else self.supergrid.q_cells()
        axes = {dim_mapping.get(axis, axis): axis for axis in var.dims}

        i_axis = axes.pop("longitude")
        i_axis = ds[i_axis].values
        j_axis = axes.pop("latitude")
        j_axis = ds[j_axis].values
        x = np.arange(i_axis.size, dtype="float")
        x_bnds = np.array([[x_ - 0.5, x_ + 0.5] for x_ in x])
        y = np.arange(j_axis.size, dtype="float")
        y_bnds = np.array([[y_ - 0.5, y_ + 0.5] for y_ in y])

        data = var.values
        lat = self.supergrid.lat
        lat_bnds = self.supergrid.lat_bnds

        lon = self.supergrid.lon
        lon_bnds = self.supergrid.lon_bnds

        # Convert time to numeric values
        time_axis = axes.pop("time")
        time_numeric = ds[time_axis].values
        time_units = ds[time_axis].attrs["units"]
        time_bnds = ds[ds[time_axis].attrs["bounds"]].values
        # TODO: Check that the calendar is the same than the one defined in the model.json
        # Convert if not.
        # calendar = ds[time_axis].attrs["calendar"]

        # CMOR setup
        ipth = "Test"
        cmor.setup(
            inpath=ipth,
            set_verbosity=cmor.CMOR_NORMAL,
            netcdf_file_action=cmor.CMOR_REPLACE,
        )

        cmor.dataset_json(cmor_dataset_json)
        current_dir = os.path.dirname(os.path.abspath(__file__))
        grid_table_id = cmor.load_table(
            os.path.join(current_dir, "cmor_tables", "CMIP6_grids.json")
        )
        cmor.set_table(grid_table_id)

        cmor_axes = []
        # Define CMOR axes
        yaxis_id = cmor.axis(
            table_entry="j_index", units="1", coord_vals=y, cell_bounds=y_bnds
        )
        xaxis_id = cmor.axis(
            table_entry="i_index", units="1", coord_vals=x, cell_bounds=x_bnds
        )

        grid_id = cmor.grid(
            axis_ids=np.array([yaxis_id, xaxis_id]),
            latitude=lat,
            longitude=lon,
            latitude_vertices=lat_bnds,
            longitude_vertices=lon_bnds,
        )
        cmor_axes.append(grid_id)

        # Now, load the Omon table to set up the time axis and variable
        current_dir = os.path.dirname(os.path.abspath(__file__))
        mip_table = os.path.join(current_dir, "cmor_tables", mip_table)
        omon_table_id = cmor.load_table(mip_table)
        cmor.set_table(omon_table_id)

        cmorTime = cmor.axis(
            "time", coord_vals=time_numeric, cell_bounds=time_bnds, units=time_units
        )
        cmor_axes.append(cmorTime)

        if axes:
            for axis, dim in axes.items():
                coord_vals = var[dim].values
                try:
                    cell_bounds = var[var[dim].attrs["bounds"]].values
                except KeyError:
                    cell_bounds = None
                axis_units = var[dim].attrs["units"]
                cmor_axis = cmor.axis(
                    axis,
                    coord_vals=coord_vals,
                    cell_bounds=cell_bounds,
                    units=axis_units,
                )
                cmor_axes.append(cmor_axis)

        # Define CMOR variable
        cmorVar = cmor.variable(cmor_name, variable_units, cmor_axes, positive=positive)

        # Write data to CMOR
        data = np.moveaxis(data, 0, -1)
        cmor.write(cmorVar, data, ntimes_passed=len(time_numeric))

        # Finalize and save the file
        filename = cmor.close(cmorVar, file_name=True)
        print("Stored in:", filename)

        cmor.close()
